/* syscall numbers needed to run as a second server on top of 4.3BSD */

#define SYS_open	5
#define SYS_fstat	62
#define SYS_access	33
#define SYS_lseek	19
#define SYS_write	4
#define SYS_read	3
#define SYS_close	6
#define SYS_ioctl	54
#define SYS_sigvec	108
#define SYS_exit	1
