/* 
 * Mach Operating System
 * Copyright (c) 1992 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon 
 * the rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	parallel.h,v $
 * Revision 2.1  92/04/21  17:16:36  rwd
 * BSDSS
 * 
 *
 */

/*
 * Serialization for threads within the kernel.
 */

extern struct mutex master_mutex;

#define master_lock() mutex_lock(&master_mutex)

#define master_unlock() mutex_unlock(&master_mutex)

#define	unix_master()	{			\
	if (p->p_master_lock++ == 0)		\
	    master_lock();			\
}

#define	unix_release()	{			\
	if (--p->p_master_lock == 0)		\
	    master_unlock();			\
}

#define	unix_reset()	{			\
	if (p->p_master_lock > 0)		\
	    p->p_master_lock = 1;		\
}

